#ifndef _TETTETTE_CPP_
#define _TETTETTE_CPP_

#include "tettette.hpp"

//#define TTT_DEBUG

// ----------------------------------------
//   Utility functions
// ----------------------------------------

// ---------- Show error message to STDERR ----------
void error(const char *fmt, ... ) {
	// from http://www.cppll.jp/cppreference/stdio_details.html#vprintf
	va_list args;
	
	va_start(args, fmt);
	vfprintf(stderr, fmt, args);
	fprintf(stderr, "\n");
	va_end(args);
}

// ---------- Get the value of a position in "vector" ----------
// (The vector is resized if "pos" is equal to or larger than vector size)
template <class TYPE> TYPE vget(vector<TYPE> &v, size_t pos){
	if(v.size() <= pos) v.resize(pos + 1, 0);
	return v[pos];
}

// ---------- Set a value to a position in "vector" ----------
// (The vector is resized if "pos" is equal to or larger than vector size)
template <class TYPE> void vstore(vector<TYPE> &v, size_t pos, TYPE val){
	if(v.size() <= pos) v.resize(pos + 1, 0);
	v[pos] = val;
}

// ---------- output UTF-16 character in UTF-8 ----------
void utf8_output(FILE * out, TTT_INT16 utf16char){
	unsigned char * tmp = (unsigned char *)&utf16char;
	// tmp[0]: lesser byte tmp[1]: greater byte
	
	if(tmp[0] <= 0x7F && tmp[1] == 0){
		fprintf(out, "%c", tmp[0]);
	}else if(tmp[1] <= 0x07){
		// 2bytes
		fprintf(out, "%c%c",
			0xC0 | ((tmp[1] & 0x07) << 2) | ((tmp[0] & 0xC0) >> 6),
			0x80 | (tmp[0] & 0x3F));
	}else{
		// 3bytes
		fprintf(out, "%c%c%c",
			0xE0 | ((tmp[1] & 0xF0) >> 4),
			0x80 | ((tmp[1] & 0x0F) << 2) | (tmp[0] >> 6),
			0x80 | (tmp[0] & 0x3F));
	}
}

// ---------- output UTF-16 character without change ----------
// (for debug)
void utf16_output(FILE * out, TTT_INT16 utf16char){
	unsigned char * tmp = (unsigned char *)&utf16char;
	// tmp[0]: lesser byte tmp[1]: greater byte
	fprintf(out, "%c%c", tmp[0], tmp[1]);
}

#define char_output utf8_output
//#define char_output utf16_output

// ---------- Print an array ----------

void print_array(FILE * out, TTT_INT16 * array){
	size_t i = 0;
	while(array[i] != 0){
		char_output(out, array[i]);
		i++;
	}
}

// ---------- Convert a decimal character to an integer ----------
TTT_INT16 char2dec(TTT_INT16 ch){
	switch(ch){
	case L'0': case FW0: return 0;
	case L'1': case FW1: return 1;
	case L'2': case FW2: return 2;
	case L'3': case FW3: return 3;
	case L'4': case FW4: return 4;
	case L'5': case FW5: return 5;
	case L'6': case FW6: return 6;
	case L'7': case FW7: return 7;
	case L'8': case FW8: return 8;
	case L'9': case FW9: return 9;
	}
	
	return ERROR_DIGIT;
}

// ---------- Convert a hexadecimal character to an integer ----------
TTT_INT16 char2hex(TTT_INT16 ch){
	TTT_INT16 val = char2dec(ch);
	if(val != ERROR_DIGIT) return val;
	
	switch(ch){
	case L'A': case L'a': case FWA: case FWa: return 10;
	case L'B': case L'b': case FWB: case FWb: return 11;
	case L'C': case L'c': case FWC: case FWc: return 12;
	case L'D': case L'd': case FWD: case FWd: return 13;
	case L'E': case L'e': case FWE: case FWe: return 14;
	case L'F': case L'f': case FWF: case FWf: return 15;
	}
	
	return ERROR_DIGIT;
}

// ---------- Convert a hexadecimal string (4 digits or less) to an integer ----------
TTT_INT16 check_charcode
(int *is_valid, TTT_INT16 c3, TTT_INT16 c2, TTT_INT16 c1, TTT_INT16 c0){
	TTT_INT16 r3, r2, r1, r0;
	
	do{
		r3 = char2hex(c3); if(r3 == ERROR_DIGIT) break;
		r2 = char2hex(c2); if(r2 == ERROR_DIGIT) break;
		r1 = char2hex(c1); if(r1 == ERROR_DIGIT) break;
		r0 = char2hex(c0); if(r0 == ERROR_DIGIT) break;
		
		// If all are valid characters as hexadecimal number
		*is_valid = 1;
		return((r3 << 12)|(r2 << 8)|(r1 << 4)|r0);
	}while(0);
	
	// If containing an invalid character
	*is_valid = 0;
	return 0;
}

// ---------- Convert a decimal string (5 digits or less) to an integer ----------
TTT_INT16 check_charcode_decimal
(int *is_valid, TTT_INT16 c4, TTT_INT16 c3, TTT_INT16 c2, TTT_INT16 c1, TTT_INT16 c0){
	TTT_INT16 r4, r3, r2, r1, r0;
	TTT_INT_COV val;
	
	do{
		r4 = char2hex(c4); if(r4 == ERROR_DIGIT) break;
		r3 = char2hex(c3); if(r3 == ERROR_DIGIT) break;
		r2 = char2hex(c2); if(r2 == ERROR_DIGIT) break;
		r1 = char2hex(c1); if(r1 == ERROR_DIGIT) break;
		r0 = char2hex(c0); if(r0 == ERROR_DIGIT) break;
		
		// If all are valid characters as decimal number
		val = r4*10000 + r3*1000 + r2*100 + r1*10 + r0;
		if(val > (TTT_INT_COV)0xFFFF) break;
		
		*is_valid = 1;
		return((TTT_INT16)val);
	}while(0);
	
	// If containing an invalid character
	*is_valid = 0;
	return 0;
}

// ----------------------------------------
//   Instance functions
// ----------------------------------------

void Tettette::build_trie(void){
	int i;
	for(i = 0; i < sizeof(TTT_Commands)/sizeof(TTT_Commands[0]); i++){
		ct.insert(TTT_Commands[i].command);
		roles.push_back(TTT_Commands[i].role);
	}
}

// ----- Parse string literal -----
int Tettette::process_string_literal(void){
	TTT_INT16 char_push;
	int flag_valid_char = 1;
	long cursor_jump_target = cursor_code; // Next cursor position if
	                                       // valid character is input
	
	if(len_cur_cmd == 0){
		// First character in the string literal or after an escape sequence
		code_prev_char = code_cur_char;
		len_cur_cmd = 1;
	}else if((code_prev_char == TE && code_cur_char == EE) ||
	         (code_prev_char == L'\'' && code_cur_char == L'"')){
		// End of buffer
		len_cur_cmd = 0;
		parsemode = TTT_Program;
	}else{
		if(code_prev_char == L'\\' && code_cur_char == L'x'){
			// ----- Escape Sequence(2-digit Hexadecimal)
			if(len_code - cursor_code >= 2){
				// No problem with position in the source code
				char_push = check_charcode(&flag_valid_char,
					L'0', L'0', code[cursor_code], code[cursor_code+1]);
				cursor_jump_target += 2;
				len_cur_cmd = 0;
			}else{
				// Otherwise
				flag_valid_char = 0;
			}
		}else if(code_prev_char == L'\\' && code_cur_char == L'u'){
			// ----- Escape Sequence(4-digit Hexadecimal)
			if(len_code - cursor_code >= 4){
				// No problem with position in the source code
				char_push = check_charcode(&flag_valid_char,
					code[cursor_code], code[cursor_code+1],
					code[cursor_code+2], code[cursor_code+3]);
				cursor_jump_target += 4;
				len_cur_cmd = 0;
			}else{
				// Otherwise
				flag_valid_char = 0;
			}
		}else if(code_prev_char == L'\\' && code_cur_char == L'd'){
			// ----- Escape Sequence(5-digit Decimal)
			if(len_code - cursor_code >= 5){
				// No problem with position in the source code
				char_push = check_charcode_decimal(&flag_valid_char,
					code[cursor_code], code[cursor_code+1], code[cursor_code+2],
					code[cursor_code+3], code[cursor_code+4]);
				cursor_jump_target += 5;
				len_cur_cmd = 0;
			}else{
				// Otherwise
				flag_valid_char = 0;
			}
		}else if(code_prev_char == L'\\'){
			// ----- Other Escape Sequences
			switch(code_cur_char){
			case L'0':  char_push = L'\0';   break;
			case L'a':  char_push = L'\a';   break;
			case L'b':  char_push = L'\b';   break;
			case L'f':  char_push = L'\f';   break;
			case L'r':  char_push = L'\x0D'; break;
			case L'n':  char_push = L'\x0A'; break;
			case L't':  char_push = L'\t';   break;
			case L'v':  char_push = L'\v';   break;
			case L'\'': char_push = L'\'';   break;
			case L'"':  char_push = L'"';    break;
			case L'\\': char_push = L'\\';   break;
			default:
				// Error
				flag_valid_char = 0;
				break;
			}
			len_cur_cmd = 0;
		}else{
			char_push = code_prev_char;
		}
		
		if(!flag_valid_char){
			error("Error: Syntax Error: Invalid escape sequence (%d)", __LINE__);
			fprintf(stderr, "Error sequence: ");
			char_output(stderr, code_prev_char);
			char_output(stderr, code_cur_char);
			fprintf(stderr, "(Position %d)\n", cursor_code - 1);
			return 2;
		}else{
			if(loop_depth == 0){
				vstore(buffer, cursor_buf, char_push);
				++cursor_buf;
			}
			cursor_code = cursor_jump_target;
			code_prev_char = code_cur_char;
		}
	}
	return 0;
}

int Tettette::process_comment(void){
 	if(code_cur_char == L'}') parsemode = TTT_Program; // End of the comment
	return 0;
}

// ----- Execute a command -----

int Tettette::process_command(int command_type){
	TTT_INT16 bufval;
	
	// ---------- Case of not skipping the program
	switch(command_type){
	case TTT_BeginBuffer:
		parsemode = TTT_Literal;
		break;
	case TTT_BeginComment:
		parsemode = TTT_Comment;
		break;
	case TTT_IncValue:
		bufval = vget(buffer, cursor_buf) + 1;
		buffer[cursor_buf] = bufval;
		break;
	case TTT_DecValue:
		bufval = vget(buffer, cursor_buf) - 1;
		buffer[cursor_buf] = bufval;
		break;
	case TTT_IncPointer:
		++cursor_buf;
		break;
	case TTT_DecPointer:
		--cursor_buf;
		break;
	case TTT_PutChar:
		char_output(stdout, buffer[cursor_buf]);
		fflush(stdout);
		break;
	case TTT_PutCharStep:
		char_output(stdout, buffer[cursor_buf]);
		fflush(stdout);
		++cursor_buf;
		break;
	case TTT_GetChar:
		if(fread(&bufval, sizeof(TTT_INT16), 1, stdin) != 1){
			error("Error: Failed in reading from STDIN (%d)", __LINE__);
			return 3;
		}
		buffer[cursor_buf] = bufval;
		break;
	case TTT_GetCharStep:
		if(fread(&bufval, sizeof(TTT_INT16), 1, stdin) != 1){
			error("Error: Failed in reading from STDIN (%d)", __LINE__);
			return 3;
		}
		buffer[cursor_buf] = bufval;
		++cursor_buf;
		break;
	case TTT_BeginLoop:
		if(vget(buffer, cursor_buf)){
			// set the beginning of this "ĂĂĂā["
			loop_pos.push(cursor_code - len_cur_cmd);
		}else{
			loop_depth = 1;
		}
		break;
	case TTT_EndLoop:
		// (loop_depth==0  TTT_EndLoop F[v̋N_ɖ߂
		if(loop_pos.empty()){
			error("Error: Loop is closed without loop beginning (%d)", __LINE__);
			return 4;
		}else{
			cursor_code = loop_pos.top();
			loop_pos.pop();
		}
		break;
	default:
		error("Error: Unexpected Error (%d)", __LINE__);
		return 5;
	}
	return 0;
}

// ----- Parse a command -----

int Tettette::process_default(void){
	int status;
	int result, role;
	
	switch(code_cur_char){
	case L'\0': case L'\t': case L'\r':
	case L'\n': case L'\f': case L' ': case SP: case BOM:
		result = -1; // Skip the character
		break;
	default:
		cmd_cur_read[len_cur_cmd] = code_cur_char;
		len_cur_cmd++;
		cmd_cur_read[len_cur_cmd] = 0;
		result = ct.find(cmd_cur_read);
	}
	
	switch(result){
	case -1: // Substring of commands
		// Do nothing
		break;
	case -2: // Not matched to any command
		error("Error: Syntax Error: Undefined command (%d)", __LINE__);
		fprintf(stderr, "Error command: ");
		for(TTT_INT16 * ptr = cmd_cur_read; *ptr; ptr++){
			char_output(stderr, *ptr);
		}
		fprintf(stderr, "\n");
		fprintf(stderr, "Error position: %d\n", cursor_code);
		return 2;
	default:
		role = roles[result];
		
		if(loop_depth == 0){
			status = process_command(role);
			if(status) return status;
		}else{
			// ---------- Case of skipping the program
			//            (because of loop exiting)
			switch(role){
			case TTT_BeginBuffer:
				parsemode = TTT_Literal;
				break;
			case TTT_BeginComment:
				parsemode = TTT_Comment;
				break;
			case TTT_IncValue: case TTT_DecValue:
			case TTT_IncPointer: case TTT_DecPointer:
			case TTT_PutChar: case TTT_GetChar:
			case TTT_PutCharStep: case TTT_GetCharStep:
				break;
			case TTT_BeginLoop:
				loop_depth++;
				break;
			case TTT_EndLoop:
				// loop_depth>0  TTT_EndLoop 
				loop_depth--;
				break;
			default:
				error("Error: Unexpected Error (%d)", __LINE__);
				return 8;
			}
		}
		
		len_cur_cmd = 0;
		cmd_cur_read[0] = 0;
		break;
	}
	return 0;
}

Tettette::Tettette(TTT_INT16 * source, long size){
	build_trie();
	len_code = size;
	exec_result = tettette_main(source);
}

int Tettette::tettette_main(TTT_INT16 * source){
	int status;
	
	memset(cmd_cur_read, '\0', MAX_COM_LEN * sizeof(TTT_INT16));
	len_cur_cmd = 0;
	cursor_code = 0;
	cursor_buf = 0;
	parsemode = TTT_Program;
	loop_depth = 0;
	
	code = source;
	
	while(cursor_code < len_code){
		code_cur_char = code[cursor_code];
		++cursor_code;
		
		if(len_cur_cmd >= MAX_COM_LEN - 1){
			fprintf(stderr, "[Error](SRC=%d PTR=%d)", len_code, cursor_code);
			print_array(stderr, cmd_cur_read);
			error("Error: Syntax Error [1] (%d)", __LINE__);
			return 1;
		}
		
		switch(parsemode){
		case TTT_Literal: // Reading String literal
			status = process_string_literal();
			if(status) return status;
			break;
		case TTT_Comment: // Reading Comment
			status = process_comment();
			if(status) return status;
			break;
		case TTT_Program: // Reading Program
			status = process_default();
			if(status) return status;
			break;
		default:
			error("Error: Unexpected Error (%d)", __LINE__);
			return 11;
		}
	}
	
	// Ending(Check validity)
	switch(parsemode){
	case TTT_Literal: // Reading String literal
		error("Error: The program ends with incomplete string literal", __LINE__);
		return 12;
	case TTT_Comment: // Reading String literal
		error("Error: The program ends with incomplete comment", __LINE__);
		return 13;
	case TTT_Program: // Reading Program
		break;
	default:
		error("Error: Unexpected Error (%d)", __LINE__);
		return 14;
	}
	if(!loop_pos.empty()){
		error("Error: The program ends with unclosed loop in position %d (%d)", loop_pos.top(), __LINE__);
		return 15;
	}
	if(loop_depth > 0){
		error("Error: The program ends with unclosed loop (%d)", __LINE__);
		return 16;
	}
	
	return 0;
}

#endif // _TETTETTE_CPP_
