#ifndef _TETTETTE_HPP_
#define _TETTETTE_HPP_

#define TETTETTE_VERSION "0.1.0"

#include "trie.hpp"
#include "get_content.hpp"
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <vector>
#include <stack>

#include "ttt-type.hpp"

// ----------------------------------------
//   Common definitions
// ----------------------------------------

// TTT_INT16 must be 2byte unsigned integer
typedef CharTrie<TTT_INT16> WCharTrie;

// UTF-16(LE)
//  U+3063 "\x63\x30" 0x3063
//  U+3066 "\x66\x30" 0x3066
// [ U+30FC "\xFC\x30" 0x30FC

#define TT 0x3063
#define TE 0x3066
#define EE 0x30FC
#define SP 0x3000 // Zenkaku Space

#define BOM 0xFEFF // Byte-Order-Mark for UTF-16LE

#define FW0 0xFF10 // Full-width Zero U+FF10 "\x10\xFF"
#define FW1 0xFF11
#define FW2 0xFF12
#define FW3 0xFF13
#define FW4 0xFF14
#define FW5 0xFF15
#define FW6 0xFF16
#define FW7 0xFF17
#define FW8 0xFF18
#define FW9 0xFF19
#define FWA 0xFF21 // Full-width A U+FF21 "\x21\xFF"
#define FWa 0xFF41 // Full-width a U+FF41 "\x41\xFF"
#define FWB 0xFF22
#define FWb 0xFF42
#define FWC 0xFF23
#define FWc 0xFF43
#define FWD 0xFF24
#define FWd 0xFF44
#define FWE 0xFF25
#define FWe 0xFF45
#define FWF 0xFF26
#define FWf 0xFF46

#define MAX_COM_LEN 8 // buffer size for one command
#define ERROR_DIGIT 255 // invalid as digit

// ----------------------------------------
//   Utility functions
// ----------------------------------------

void showhelp(void);
void error(const char *fmt, ... );
void utf8output(FILE * out, TTT_INT16 utf16char);
void print_array(FILE * out, TTT_INT16 * array);

// get/store a value of a vector, with expanding the vector if specified
// position exceeds the current size
template <class TYPE> TYPE vget(vector<TYPE> &v, size_t pos);
template <class TYPE> void vstore(vector<TYPE> &v, size_t pos, TYPE val);

// ----------------------------------------
//   Class definitions
// ----------------------------------------

// R}h̎
typedef enum _TTT_CommandRole{
	TTT_BeginBuffer, // [             ` (Beginning of buffer; end with '")
	TTT_BeginComment,//                { (Beginning of comment; end with })
	TTT_IncValue,    // Ăā[         + (Increase value of pointed buffer)
	TTT_DecValue,    // Ă[         - (Decrease value of pointed buffer)
	TTT_IncPointer,  // Ăā[       > (Increase pointer)
	TTT_DecPointer,  // ĂĂā[     < (Decrease pointer)
	TTT_PutChar,     //                . (putc)
	TTT_GetChar,     //                , (getc)
	TTT_PutCharStep, // ĂĂ[     ) (putc and increase pointer for one)
	TTT_GetCharStep, // ĂĂā[   ( (getc and increase pointer for one)
	TTT_BeginLoop,   // ĂĂĂā[ [ (beginning of a loop)
	TTT_EndLoop,     // ĂĂĂ[ ] (end of a loop)
} TTT_CommandRole;

// ݃\[XR[hŌĂʒu͉ӖĂ邩
typedef enum _TTT_ParseMode{
	TTT_Program, // ʏ̃vO
	TTT_Literal, // 񃊃e
	TTT_Comment, // Rg
} TTT_ParseMode;

typedef struct _TTT_CommandInfo{
	TTT_CommandRole role;
	TTT_INT16 command[MAX_COM_LEN];
} TTT_Command_Info;

const TTT_Command_Info TTT_Commands[] = {
	{TTT_BeginBuffer, {EE,0}},
	{TTT_IncValue,    {TE,TE,EE,0}},
	{TTT_DecValue,    {TE,TT,EE,0}},
	{TTT_IncPointer,  {TE,TT,TE,EE,0}},
	{TTT_DecPointer,  {TE,TT,TE,TE,EE,0}},
	{TTT_GetCharStep, {TE,TT,TE,TT,TE,EE,0}},
	{TTT_PutCharStep, {TE,TT,TE,TT,EE,0}},
	{TTT_BeginLoop,   {TE,TT,TE,TT,TE,TE,EE,0}},
	{TTT_EndLoop,     {TE,TT,TE,TT,TE,TT,EE,0}},
	
	{TTT_BeginBuffer, {L'`',0}},
	{TTT_BeginComment,{L'{',0}},
	{TTT_IncValue,    {L'+',0}},
	{TTT_DecValue,    {L'-',0}},
	{TTT_IncPointer,  {L'>',0}},
	{TTT_DecPointer,  {L'<',0}},
	{TTT_GetChar,     {L',',0}},
	{TTT_PutChar,     {L'.',0}},
	{TTT_GetCharStep, {L'(',0}},
	{TTT_PutCharStep, {L')',0}},
	{TTT_BeginLoop,   {L'[',0}},
	{TTT_EndLoop,     {L']',0}},
};
	
class Tettette{
private:
	// ---------- variables
	WCharTrie ct; // Trie for commands
	vector<TTT_CommandRole> roles;
	
	TTT_INT16 * code; // source code
	long len_code; // Size of the source code
	
	TTT_INT16 cmd_cur_read[MAX_COM_LEN]; // Substring from the position next to
	                                     // last command to current cursor
	int len_cur_cmd; // String length read in `cmd_cur_read'
	
	vector<TTT_INT16> buffer; // working buffer for user
	long cursor_code; // Position in the source code to be parsed now
	long cursor_buf; // Pointer managed by the "Tettette" program
	TTT_INT16 code_cur_char; // Character in the source code parsed now
	TTT_INT16 code_prev_char; // Character in the source code parsed one time before
	                          // (Supporting variable for parsing string literal)
	stack<long> loop_pos; // Positions in the source code storing initial
	                      // points of loops
	
	TTT_ParseMode parsemode; // 0: default, -1: string literal
	int loop_depth;
	
	// ---------- instance functions
	void build_trie(void);
	int process_string_literal(void);
	int process_comment(void);
	int process_command(int command_type);
	int process_default(void);
	
public:
	// ---------- variables
	int exec_result; // status returned to parent process
	
	// ---------- instance functions
	Tettette(TTT_INT16 * source, long size);
	int tettette_main(TTT_INT16 * source);
};

//#include "tettette.cpp"

#endif // _TETTETTE_HPP_
