#ifndef _TRIE_CPP_
#define _TRIE_CPP_

#include "trie.hpp"

// ---------- Functions for TrieNodeVA ----------

template <class NODETYPE>
NODETYPE & TrieNodeVA<NODETYPE>::val(void){
	return nodeval;
}

template <class NODETYPE>
typename TrieNodeVA<NODETYPE>::ChildItem TrieNodeVA<NODETYPE>::seek(size_t index){
	return addchild(index, 1);
}

template <class NODETYPE>
typename TrieNodeVA<NODETYPE>::ChildItem TrieNodeVA<NODETYPE>::addchild(size_t index, int onlyseek = 0){
	int min = 0;
	int max = children.size() - 1;
	int pos;
	TrieNodeVA<NODETYPE> * newdest;
	NodeIndex tmp;
	
	if(children.size() > 0){
		// binary search
		while(min <= max){
			pos = (max + min) / 2;
			if(children[pos].index == index){
				return (children.begin() + pos);
			}else if(children[pos].index > index){
				max = pos - 1;
			}else{
				min = pos + 1;
			}
		}
		
		if(onlyseek) return children.end();
		
		if(children[pos].index < index) pos++;
	}else{
		if(onlyseek) return children.end();
		pos = 0;
	}
	
	newdest = new TrieNodeVA<NODETYPE>;
	tmp.index = index;
	tmp.dest = newdest;
	return children.insert(children.begin() + pos, tmp);
}

template <class NODETYPE>
TrieNodeVA<NODETYPE>::TrieNodeVA(void){
	// nothing to do
}

template <class NODETYPE>
int TrieNodeVA<NODETYPE>::link(TrieNodeVA<NODETYPE> * other, size_t index){
	ChildItem cp = seek(index);
	if(cp != children.end()) return 0;
	cp = addchild(index, 0);
	cp->index = index;
	cp->dest = other;
	return 1;
}

template <class NODETYPE>
int TrieNodeVA<NODETYPE>::unlink(size_t index){
	ChildItem cp = seek(index);
	if(cp == children.end()) return 0;
	children.erase(cp);
	return 1;
}

template <class NODETYPE>
TrieNodeVA<NODETYPE> * TrieNodeVA<NODETYPE>::child(size_t index){
	ChildItem cp = seek(index);
	if(cp == children.end()) return NULL;
	return cp->dest;
}

// ---------- Functions for CharTrie ----------

template <class TYPE> CharTrie<TYPE>::CharTrie(void){
	CharTrieNodeVA * ctn = new CharTrieNodeVA;
	ctn->val() = -1;
	nodes.push_back(ctn);
	
	node_id = 0;
}

template <class TYPE> CharTrie<TYPE>::~CharTrie(){
	list<CharTrieNodeVA *>::iterator it;
	while(!nodes.empty()){
		delete *(nodes.begin());
		nodes.pop_front();
	}
}

template <class TYPE> int CharTrie<TYPE>::track_nodes
(const TYPE *str, typename CharTrie<TYPE>::TrackNodeMode mode){
	CharTrieNodeVA * ctn = *(nodes.begin());
	CharTrieNodeVA * tmp;
	size_t i = 0;
	
	while(str[i] != 0){
		tmp = ctn->child(str[i]);
		if(tmp == NULL) break;
		
		ctn = tmp;
		i++;
	}
	
	if(str[i] == 0){
		switch(mode){
		case tnm_insert:
			// Specified string is already stored in the trie
			if(ctn->val() >= 0) return 0;
			
			// Already stored but not end of the word
			ctn->val() = (node_id++);
			return 1;
			
			break;
		case tnm_find:
			// Specified string is already stored in the trie
			if(ctn->val() >= 0) return ctn->val();
			
			// Already stored but not end of the word
			return -1;
			
			break;
		default:
			fprintf(stderr, "TRIE library error\n");
			exit(-1);
		}
	}
	
	// Not stored in the TRIE
	switch(mode){
	case tnm_find:
		return -2;
	}
	
	// Store a new string
	while(str[i] != 0){
		tmp = new CharTrieNodeVA;
		nodes.push_back(tmp);
		
		ctn->link(tmp, str[i]);
		tmp->val() = -1;
		
		ctn = tmp;
		i++;
	}
	ctn->val() = (node_id++);
	
	return 1;
}

// 1 if `str` is newly inserted, 0 otherwise
template <class TYPE> int CharTrie<TYPE>::insert(const TYPE *str){
	return track_nodes(str, tnm_insert);
}

// 0 if `str` is in the TRIE, 2 if `str` is in the tree but substring,
// 1 otherwise
template <class TYPE> int CharTrie<TYPE>::find(const TYPE *str){
	return track_nodes(str, tnm_find);
}

// number of nodes, not strings
template <class TYPE> int CharTrie<TYPE>::count(void){
	return nodes.size();
}

// number of strings
template <class TYPE> int CharTrie<TYPE>::size(void){
	return node_id;
}

#endif // _TRIE_CPP_
